function MapExtension.GUI.CreateFrame(parent, align, x, y, w, h, layer, uiParent)
	local frame = UI.CreateFrame("Frame", "", iif(uiParent == nil, parent, uiParent))
	frame:SetPoint(align, parent, "TOPLEFT", x, y)
	frame:SetWidth(w)
	frame:SetHeight(h)
	frame:SetLayer(layer)
	return frame
end

function MapExtension.GUI.CreateText(parent, align, x, y, text, layer, size, color)
	local frame = UI.CreateFrame("Text", "", parent)
	frame:SetFontSize(size or 14)
	frame:SetPoint(align, parent, align, x, y)
	if text then frame:SetText(text) end
	if layer then frame:SetLayer(layer) end
	if color then frame:SetFontColor(color[1], color[2], color[3], color[4]) end

	return frame
end

function MapExtension.GUI.CreateRiftButton(parent, x, y, layer)
	local frame = UI.CreateFrame("RiftButton", "", parent)
	frame:SetPoint("TOPRIGHT", parent, "TOPRIGHT", x, y)
	frame:SetSkin("close")
	frame:SetLayer(layer)
	return frame
end

function MapExtension.GUI.CreateButton(parent, x, y, w, h, layer, texture, layout, isSwitch)
	local frame = UI.CreateFrame("Texture", "", parent)
	if x and y then frame:SetPoint(ifnull(layout, "TOPLEFT"), parent, ifnull(layout, "TOPLEFT"), x, y) end
	frame:SetWidth(w)
	frame:SetHeight(h)
	frame:SetLayer(layer)
	frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-normal.png")
	frame.state = false

	function frame.Event:LeftDown()
		frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-down.png")
	end

	function frame.Event:MouseIn()
		frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-" .. iif(isSwitch and frame.state, "down", "") .. "hover.png")
	end

	function frame.Event:MouseMove()
		frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-" .. iif(isSwitch and frame.state, "down", "") .. "hover.png")
	end

	function frame.Event:MouseOut()
		frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-" .. iif(isSwitch and frame.state, "down", "normal") .. ".png")
	end

	function frame.SetState(flag)
		frame.state = flag
		frame:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. "-" .. iif(isSwitch and frame.state, "down", "normal") .. ".png")
	end
	return frame
end

function MapExtension.GUI.CreateCheckbox(parent, align, x, y, layer, text, callback, param, value, iconTexture)
	local frame = MapExtension.GUI.CreateFrame(parent, align, x, y, 100, 19, 0)

	frame.checkbox = UI.CreateFrame("RiftCheckbox", "", frame)
	frame.checkbox:SetPoint("TOPLEFT", frame, "TOPLEFT", 7, 2)
	frame.checkbox:SetLayer(layer)
	frame.checkbox:SetEnabled(true)
	frame.checkbox:SetChecked(ifnull(value, false))

	if iconTexture ~= nil then
		MapExtension.GUI.CreateSimpleTexture(frame, 24, 2, 17, 17, layer):SetTexture(MapExtensionAddon.identifier, iconTexture)
	end

	frame.text = UI.CreateFrame("Text", "", frame)
	frame.text:SetPoint("TOPLEFT", frame, "TOPLEFT", iif(iconTexture == nil, 23, 45), 0)
	frame.text:SetFontSize(14)
	frame.text:SetText(text)
	
	frame.checkbox.Event.CheckboxChange = function() callback(frame, param) end
	
	return frame
end

function MapExtension.GUI.CreateHoveredCheckbox(parent, align, x, y, width, height, layer, text, callbackChange, callbackHover, param, value)
	local frame = MapExtension.GUI.CreateFrame(parent, align, x, y, width, height, 0)

	frame.checkbox = UI.CreateFrame("RiftCheckbox", "", frame)
	frame.checkbox:SetPoint("TOPLEFT", frame, "TOPLEFT", 7, 2)
	frame.checkbox:SetLayer(layer)
	frame.checkbox:SetEnabled(true)
	frame.checkbox:SetChecked(ifnull(value, false))
	
	frame.text = UI.CreateFrame("Text", "", frame)
	frame.text:SetPoint("TOPLEFT", frame, "TOPLEFT", 23, 0)
	frame.text:SetFontSize(14)
	frame.text:SetText(text)

	MapExtension.GUI.CreateSimpleTexture(frame, -5, 2, 9, 15, 100, "img/folder.png", "TOPRIGHT")
        frame.checkbox.Event.CheckboxChange = function() callbackChange(frame, param) end
	frame.Event.MouseIn = function() callbackHover(frame, param) end
	
	function frame.SetHighlight(flag)
		frame:SetBackgroundColor(1, 1, 1, iif(flag, .1, 0)) 
	end

	return frame
end

function MapExtension.GUI.CreateTileTexture(parent, marginTop, marginRight, marginBottom, marginLeft, w, h, layer, textureC, textureH, textureV)
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "TOPLEFT", x = marginLeft, y = marginTop }}, layer, textureC, w, h)
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "TOPRIGHT", x = -marginRight, y = marginTop }}, layer, textureC, w, h)
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "BOTTOMLEFT", x = marginLeft, y = -marginBottom }}, layer, textureC, w, h)
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "BOTTOMRIGHT", x = -marginRight, y = -marginBottom }}, layer, textureC, w, h)
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "TOPLEFT", x = marginLeft + w, y = marginTop }, { layout = "TOPRIGHT", x = - marginRight - w, y = marginTop }}, layer, textureH, nil, h, {{ layout = "TOPLEFT", x = 0, y = 0 }, { layout = "TOPRIGHT", x = 0, y = 0 }})
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "BOTTOMLEFT", x = marginLeft + w, y = -marginBottom }, { layout = "BOTTOMRIGHT", x = - marginRight - w, y = -marginBottom }}, layer, textureH, nil, h, {{ layout = "BOTTOMLEFT", x = 0, y = 0 }, { layout = "BOTTOMRIGHT", x = 0, y = 0 }})
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "TOPLEFT", x = marginLeft, y = marginTop + h }, { layout = "BOTTOMLEFT", x = marginLeft, y = -marginBottom - h }}, layer, textureV, w, nil, {{ layout = "TOPLEFT", x = 0, y = 0 }, { layout = "BOTTOMLEFT", x = 0, y = 0 }})
	MapExtension.GUI.CreateTexture(MapExtension.AE.UI.Window, {{ layout = "TOPRIGHT", x = -marginRight, y = marginTop + h }, { layout = "BOTTOMRIGHT", x = -marginRight, y = -marginBottom - h }}, layer, textureV, w, nil, {{ layout = "TOPRIGHT", x = 0, y = 0 }, { layout = "BOTTOMRIGHT", x = 0, y = 0 }})
end

function MapExtension.GUI.CreateTexture(parent, layout, layer, texture, w, h, layoutTexture)
	local frame = UI.CreateFrame("Mask", "", parent)
	for idx, layoutItem in ipairs(layout) do
		frame:SetPoint(layoutItem.layout, parent, layoutItem.layout, layoutItem.x, layoutItem.y)
	end
	if w then frame:SetWidth(w) end
	if h then frame:SetHeight(h) end
	frame:SetLayer(layer)

	frame.texture = UI.CreateFrame("Texture", "", frame)
	frame.texture:SetTexture(MapExtensionAddon.identifier, texture)

	if layoutTexture == nil then layoutTexture = layout end
	for idx, layoutItem in ipairs(layoutTexture) do
		frame.texture:SetPoint(layoutItem.layout, frame, layoutItem.layout, 0, 0)
	end
	return frame
end

function MapExtension.GUI.CreateSimpleTexture(parent, x, y, w, h, layer, texture, layout)
	local frame = UI.CreateFrame("Texture", "", parent)
	frame:SetPoint(ifnull(layout, "TOPLEFT"), parent, ifnull(layout, "TOPLEFT"), x, y)
	frame:SetWidth(w)
	frame:SetHeight(h)
	frame:SetLayer(layer)
	if texture then	frame:SetTexture(MapExtensionAddon.identifier, texture) end
	return frame
end

function MapExtension.GUI.CreateTextureBlock(id, parent, texture, x, y, level)
	if level == 0 then
		MapExtension.GUI.CreateTextureSimple(id, parent, {{ layout1 = "TOPLEFT", layout2 = "TOPLEFT" }, { layout1 = "BOTTOMRIGHT", layout2 = "CENTER" }}, texture .. string.sub("0" .. y, -2) .. "." .. string.sub("0" .. x, -2) .. ".jpg")
		MapExtension.GUI.CreateTextureSimple(id, parent, {{ layout1 = "BOTTOMLEFT", layout2 = "CENTER" }, { layout1 = "TOPRIGHT", layout2 = "TOPRIGHT" }}, texture .. string.sub("0" .. y, -2) .. "." .. string.sub("0" .. (x + 1), -2) .. ".jpg")
		MapExtension.GUI.CreateTextureSimple(id, parent, {{ layout1 = "BOTTOMLEFT", layout2 = "BOTTOMLEFT" }, { layout1 = "TOPRIGHT", layout2 = "CENTER" }}, texture .. string.sub("0" .. (y + 1), -2) .. "." .. string.sub("0" .. x, -2) .. ".jpg")
		MapExtension.GUI.CreateTextureSimple(id, parent, {{ layout1 = "TOPLEFT", layout2 = "CENTER" }, { layout1 = "BOTTOMRIGHT", layout2 = "BOTTOMRIGHT" }}, texture .. string.sub("0" .. (y + 1), -2) .. "." .. string.sub("0" .. (x + 1), -2) .. ".jpg")
	else
		MapExtension.GUI.CreateTextureBlock(id, MapExtension.GUI.CreateFrameSimple(parent, {{ layout1 = "TOPLEFT", layout2 = "TOPLEFT" }, { layout1 = "BOTTOMRIGHT", layout2 = "CENTER" }}), texture, x, y, level - 1)
		MapExtension.GUI.CreateTextureBlock(id, MapExtension.GUI.CreateFrameSimple(parent, {{ layout1 = "BOTTOMLEFT", layout2 = "CENTER" }, { layout1 = "TOPRIGHT", layout2 = "TOPRIGHT" }}), texture, x + math.pow(2, level), y, level - 1)
		MapExtension.GUI.CreateTextureBlock(id, MapExtension.GUI.CreateFrameSimple(parent, {{ layout1 = "BOTTOMLEFT", layout2 = "BOTTOMLEFT" }, { layout1 = "TOPRIGHT", layout2 = "CENTER" }}), texture, x, y + math.pow(2, level), level - 1)
		MapExtension.GUI.CreateTextureBlock(id, MapExtension.GUI.CreateFrameSimple(parent, {{ layout1 = "TOPLEFT", layout2 = "CENTER" }, { layout1 = "BOTTOMRIGHT", layout2 = "BOTTOMRIGHT" }}), texture, x + math.pow(2, level), y + math.pow(2, level), level - 1)
	end
end

function MapExtension.GUI.CreateTextureSimple(id, parent, layout, texture)
	local frame = UI.CreateFrame("Texture", "", parent)
	for idx, layoutItem in ipairs(layout) do
		frame:SetPoint(layoutItem.layout1, parent, layoutItem.layout2, iif(layoutItem.layout2 == "CENTER", iif(layoutItem.layout1 == "BOTTOMRIGHT" or layoutItem.layout1 == "TOPRIGHT", 1, 0), 0), iif(layoutItem.layout2 == "CENTER", iif(layoutItem.layout1 == "BOTTOMRIGHT" or layoutItem.layout1 == "BOTTOMLEFT", 1, 0), 0))
	end
	MapExtension.EGUI.SetTextureAsync(id, frame, MapExtensionAddon.identifier, texture)
	return frame
end

function MapExtension.GUI.CreateFrameSimple(parent, layout)
	local frame = UI.CreateFrame("Frame", "", parent)
	for idx, layoutItem in ipairs(layout) do
		frame:SetPoint(layoutItem.layout1, parent, layoutItem.layout2, iif(layoutItem.layout2 == "CENTER", iif(layoutItem.layout1 == "BOTTOMRIGHT" or layoutItem.layout1 == "TOPRIGHT", 1, 0), 0), iif(layoutItem.layout2 == "CENTER", iif(layoutItem.layout1 == "BOTTOMRIGHT" or layoutItem.layout1 == "BOTTOMLEFT", 1, 0), 0))
	end
	return frame
end

function MapExtension.GUI.CreateTextureView(parent, id, marginTop, marginRight, marginBottom, marginLeft, layer, texture, width, height, borderWidth, offset)
	local frame = UI.CreateFrame("Frame", "", parent)
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", marginLeft, marginTop)
	frame:SetPoint("BOTTOMRIGHT", parent, "BOTTOMRIGHT", -marginRight, -marginBottom)
	frame:SetLayer(layer)

	frame.zoom = 100.0
	frame.scale = 1.2
	frame.textureSize = { width = width, height = height, ratio = width / height, offset = offset }
	frame.offset = { x = 0, y = 0 }
	frame.activeOffset = { x = 0, y = 0 }
	frame.diff = { x = 0, y = 0 }

	frame.MousePosition = {}
	frame.MouseLeftClick = {}
	frame.MouseRightClick = {}
	frame.Marks = {}

	frame.mask = UI.CreateFrame("Mask", "", frame)
	frame.mask:SetAllPoints(frame)

	frame.tooltip = MapExtension.Tooltip.Init(parent)

	frame.texture = UI.CreateFrame("Frame", "", frame.mask)
	frame.texture:SetPoint("CENTER", frame.mask, "CENTER", 0, 0)
	frame.texture:SetLayer(layer)
	frame.texture:SetWidth(frame.textureSize.width)
	frame.texture:SetHeight(frame.textureSize.height)

	frame.field = UI.CreateFrame("Frame", "", frame.texture)
	frame.field:SetPoint("TOPLEFT", frame.texture, "TOPLEFT", 0, 0)
	frame.field:SetPoint("BOTTOMRIGHT", frame.texture, "BOTTOMRIGHT", 0, 0)

	MapExtension.GUI.CreateTextureBlock(id, frame.field, texture, 0, 0, 3)

	function frame.Event:WheelForward()
		if self.zoom > 1 then
			local mouse = Inspect.Mouse()
			self.offset.x = self.offset.x - (mouse.x - self:GetLeft() - self:GetWidth() / 2) * self.zoom * (frame.scale - 1)
			self.offset.y = self.offset.y - (mouse.y - self:GetTop() - self:GetHeight() / 2) * self.zoom * (frame.scale - 1)
			self.zoom = self.zoom / frame.scale
			self:Update()
			self:OnMousePosition()
		end
	end

	function frame.Event:WheelBack()
		local mouse = Inspect.Mouse()
		self.offset.x = self.offset.x + (mouse.x - self:GetLeft() - self:GetWidth() / 2) * self.zoom * (frame.scale - 1)
		self.offset.y = self.offset.y + (mouse.y - self:GetTop() - self:GetHeight() / 2) * self.zoom * (frame.scale - 1)
		self.zoom = self.zoom * frame.scale
		self:Update()
		self:OnMousePosition()
	end
	
	function frame.Event:Size()
		self:Update()
		self:OnMousePosition()
	end

	function frame.Event:LeftDown()
		frame.mousePosition = Inspect.Mouse()
	end

	function frame.Event:LeftUp()
		frame.offset = { x = frame.offset.x + frame.diff.x, y = frame.offset.y + frame.diff.y }
		frame.diff = { x = 0, y = 0 }
		frame.mousePosition = nil
		frame:Update()
	end

	frame.Event.LeftUpoutside = frame.Event.LeftUp

	function frame.Event:MouseMove()
		frame.isHover = true
		if frame.mousePosition then
			local mouse = Inspect.Mouse()
			frame.diff = { x = (mouse.x - frame.mousePosition.x) * frame.zoom, 
				       y = (mouse.y - frame.mousePosition.y) * frame.zoom }
			frame:Update()
		end
		frame:OnMousePosition()
	end

	function frame.Event:MouseOut()
		frame.isHover = false
		frame:OnMousePosition()
	end

	function frame:OnMousePosition()
		local mouse = Inspect.Mouse()
		for idx, eventItem in ipairs(frame.MousePosition) do
			eventItem(frame.isHover, (mouse.x - frame:GetLeft() - frame:GetWidth() / 2) * frame.zoom - frame.activeOffset.x + frame.textureSize.offset.x,
						(mouse.y - frame:GetTop() - frame:GetHeight() / 2) * frame.zoom - frame.activeOffset.y + frame.textureSize.offset.y)
		end
	end

	function frame:SetPosition(x, y)
		self.offset = { x = frame.textureSize.offset.x - x, y = frame.textureSize.offset.y - y }
		self:Update()
	end

	function frame:Update()
		self.zoom = math.min(math.max(1, self.zoom), math.max(self.textureSize.width / self:GetWidth(), self.textureSize.height / self:GetHeight()))
		if self.zoom ~= self.oldzoom then
			self.oldzoom = self.zoom
			self:UpdateMarks()
			MapExtension.Tooltip.Hide()
		end

		local mx = math.max(math.max((self.textureSize.width - self:GetWidth() * self.zoom) / 2, 0) - borderWidth, 0)
		local my = math.max(math.max((self.textureSize.height - self:GetHeight() * self.zoom) / 2, 0) - borderWidth, 0)
		self.offset = { x = math.min(math.max(self.offset.x, -mx), mx), y = math.min(math.max(self.offset.y, -my), my) }
		self.activeOffset = { x = math.min(math.max(self.offset.x + self.diff.x, -mx), mx), y = math.min(math.max(self.offset.y + self.diff.y, -my), my) }

		self.texture:SetPoint("CENTER", self.mask, "CENTER", self.activeOffset.x / self.zoom, self.activeOffset.y / self.zoom)
		self.texture:SetWidth(self.textureSize.width / self.zoom)
		self.texture:SetHeight(self.textureSize.height / self.zoom)
	end

	function frame:AddMark(name, texture, position, data, flag)
		local mark = UI.CreateFrame("Texture", "", self.texture)
		if flag ~= nil then mark:SetVisible(flag) end
		mark.position = position
		mark.data = data
		mark:SetLayer(position.layer or 100)
		mark:SetTexture(MapExtensionAddon.identifier, "img/" .. texture .. ".png")

		mark.Event.LeftDown = frame.Event.LeftDown
		mark.Event.LeftUp = frame.Event.LeftUp
		mark.Event.LeftUpoutside = frame.Event.LeftUp
		mark.Event.MouseMove = frame.Event.MouseMove
		function mark.Event:MouseIn() mark.hover = true mark.Show() end
		function mark.Event:MouseMove() mark.hover = true mark.Show() end
		function mark.Event:MouseOut() frame.Event:MouseOut() mark.hover = false mark.Show() end
		function mark.Show() if mark.data then if mark.hover then MapExtension.Tooltip.Show(name, mark) else MapExtension.Tooltip.Hide(name) end end end
		function mark.ChangeTexture(source, texture, width, height)
			if mark.textureSource == nil then 
				mark.textureSource, mark.textureName = mark:GetTexture() 
				mark.textureWidth = mark:GetWidth()
				mark.textureHeight = mark:GetHeight()
				mark.textureLayer = mark:GetLayer()
			end
			mark:SetTexture(source or mark.textureSource, texture or mark.textureName)
			mark:SetWidth(width or mark.textureWidth)
			mark:SetHeight(height or mark.textureHeight)
			mark:SetLayer(iif(source == nil, mark.textureLayer, mark.textureLayer + 100))
		end

		if self.Marks[name] == nil then self.Marks[name] = {} end
		table.insert(self.Marks[name], mark)
		mark:SetPoint("CENTER", self.texture, "CENTER", (position.x - frame.textureSize.offset.x) / self.zoom, (position.y - frame.textureSize.offset.y) / self.zoom)
		return mark
	end

	function frame:SetMarkProperties(name, position, caption)
		if self.Marks[name] then
			for k, v in ipairs(self.Marks[name]) do
				v.data.caption = caption
				v.position = position
				v:SetPoint("CENTER", self.texture, "CENTER", (position.x - frame.textureSize.offset.x) / self.zoom, (position.y - frame.textureSize.offset.y) / self.zoom)
			end
		end
	end

	function frame:UpdateMarks()
		for k, v in pairs(self.Marks) do
			for mk, mv in ipairs(v) do
				mv:SetPoint("CENTER", self.texture, "CENTER", (mv.position.x - frame.textureSize.offset.x) / self.zoom, (mv.position.y - frame.textureSize.offset.y) / self.zoom)
			end
		end
	end

	function frame:SetVisibleMark(name, flag)
		if self.Marks[name] then
			for k, v in ipairs(self.Marks[name]) do
				v:SetVisible(flag)
			end
		end
	end

	frame:Update()
	return frame
end
